@@ -32,8 +32,8 @@
 
 local _VERSION = _VERSION;
 local MAJOR, MINOR = assert(_VERSION:match "^Lua (%d+).(%d+)$");
-if tonumber(MAJOR.."."..MINOR) < 5.2 then
-  error "NSE requires Lua 5.2 or newer. It looks like you're using an older version of nmap."
+if tonumber(MAJOR.."."..MINOR) < 5.3 then
+  error "NSE requires Lua 5.3 or newer. It looks like you're using an older version of nmap."
 end
 
 local NAME = "NSE";
@@ -51,6 +51,7 @@ local DESTRUCTOR = "NSE_DESTRUCTOR";
 local SELECTED_BY_NAME = "NSE_SELECTED_BY_NAME";
 local FORMAT_TABLE = "NSE_FORMAT_TABLE";
 local FORMAT_XML = "NSE_FORMAT_XML";
+local PARALLELISM = "NSE_PARALLELISM";
 
 -- Unique value indicating the action function is going to run.
 local ACTION_STARTING = {};
@@ -128,13 +129,17 @@ local remove = table.remove;
 local sort = table.sort;
 local unpack = table.unpack;
 
+local os = require "os"
+local time = os.time
+local difftime = os.difftime
+
 do -- Add loader to look in nselib/?.lua (nselib/ can be in multiple places)
   local function loader (lib)
     lib = lib:gsub("%.", "/"); -- change Lua "module separator" to directory separator
     local name = "nselib/"..lib..".lua";
     local type, path = cnse.fetchfile_absolute(name);
     if type == "file" then
-      return loadfile(path);
+      return assert(loadfile(path));
     else
       return "\n\tNSE failed to find "..name.." in search paths.";
     end
@@ -234,6 +239,11 @@ local function log_error (fmt, ...)
   log_write("stderr", format(fmt, ...));
 end
 
+-- Check for and warn about some known bad behaviors
+if ("test"):gsub(".*$", "x") == "xx" then
+  log_error("Known bug in string.gsub in Lua 5.3 before 5.3.3 will cause bugs in NSE scripts.")
+end
+
 local function table_size (t)
   local n = 0; for _ in pairs(t) do n = n + 1; end return n;
 end
@@ -256,17 +266,7 @@ end
 
 -- recursively copy a table, for host/port tables
 -- not very rigorous, but it doesn't need to be
-local function tcopy (t)
-  local tc = {};
-  for k,v in pairs(t) do
-    if type(v) == "table" then
-      tc[k] = tcopy(v);
-    else
-      tc[k] = v;
-    end
-  end
-  return tc;
-end
+local tcopy = require "tableaux".tcopy
 
 -- copies the host table while preserving the registry
 local function host_copy(t)
@@ -384,10 +384,16 @@ do
   -- prerule/postrule scripts may be timed out in the future
   -- based on start time and script lifetime?
   function Thread:timed_out ()
-    if self.type == "hostrule" or self.type == "portrule" then
-      return cnse.timedOut(self.host);
+    local host_timeout, script_timeout = false, false;
+    -- checking whether user gave --script-timeout option or not
+    if cnse.script_timeout and cnse.script_timeout > 0 then
+      -- comparing script's timeout with time elapsed
+      script_timeout = cnse.script_timeout < difftime(time(), self.start_time)
     end
-    return nil;
+    if self.type == "hostrule" or self.type == "portrule" then
+      host_timeout = cnse.timedOut(self.host);
+    end
+    return script_timeout or host_timeout;
   end
 
   function Thread:start_time_out_clock ()
@@ -408,6 +414,12 @@ do
       timeouts[self.host] = timeouts[self.host] or {};
       timeouts[self.host][self.co] = true;
     end
+    -- storing script's start time so as to account for script's timeout later
+    if self.worker then
+      self.start_time = self.parent.start_time
+    else
+      self.start_time = time()
+    end
   end
 
   -- Remove scripts from the timeouts list and call their
@@ -473,6 +485,7 @@ do
       script = self,
       type = script_type,
       worker = false,
+      start_time = 0, --for script timeout
     };
     thread.parent = thread;
     setmetatable(thread, Thread)
@@ -489,6 +502,7 @@ do
       info = format("%s W:%s", self.id, match(tostring(co), "^thread: 0?[xX]?(.*)"));
       parent = self,
       worker = true,
+      start_time = 0,
     };
     setmetatable(thread, Worker)
     local function info ()
@@ -704,13 +718,14 @@ local function get_chosen_scripts (rules)
 
   for i, rule in ipairs(rules) do
     rule = match(rule, "^%s*(.-)%s*$"); -- strip surrounding whitespace
-    local original_rule = rule;
     local forced, rule = is_forced_set(rule);
-    used_rules[rule] = false; -- has not been used yet
-    forced_rules[rule] = forced;
-    -- Here we escape backslashes which might appear in Windows filenames.
-    rule = gsub(rule, "\\([^\\])", "\\\\%1");
-    rules[i] = rule;
+    if rule and rule ~= "" then
+      used_rules[rule] = false; -- has not been used yet
+      forced_rules[rule] = forced;
+      -- Here we escape backslashes which might appear in Windows filenames.
+      rule = gsub(rule, "\\([^\\])", "\\\\%1");
+      rules[i] = rule;
+    end
   end
 
   -- Checks if a given script, script_entry, should be loaded. A script_entry
@@ -798,7 +813,10 @@ local function get_chosen_scripts (rules)
         t, path = cnse.fetchscript(rule..".nse");
       end
       if t == nil then
-        error("'"..rule.."' did not match a category, filename, or directory");
+        -- Avoid erroring if -sV but no scripts are present
+        if not (cnse.scriptversion and rule == "version") then
+          error("'"..rule.."' did not match a category, filename, or directory");
+        end
       elseif t == "file" and not files_loaded[path] then
         script_params.selection = "file path";
         script_params.verbosity = true;
@@ -822,7 +840,8 @@ local function get_chosen_scripts (rules)
   -- calculate runlevels
   local name_script = {};
   for i, script in ipairs(chosen_scripts) do
-    assert(name_script[script.short_basename] == nil);
+    assert(name_script[script.short_basename] == nil,
+      ("duplicate script ID: '%s'"):format(script.short_basename));
     name_script[script.short_basename] = script;
   end
   local chain = {}; -- chain of script names
@@ -958,7 +977,8 @@ local function run (threads_iter, hosts)
     end
 
     local nr, nw = table_size(running), table_size(waiting);
-    if cnse.key_was_pressed() then
+    -- total may be 0 if no scripts are running in this phase
+    if total > 0 and cnse.key_was_pressed() then
       print_verbose(1, "Active NSE Script Threads: %d (%d waiting)",
           nr+nw, nw);
       progress("printStats", 1-(nr+nw)/total);
@@ -971,8 +991,43 @@ local function run (threads_iter, hosts)
           thread:d("Waiting: %THREAD_AGAINST\n\t%s",
               (gsub(traceback(co), "\n", "\n\t")));
         end
+      elseif debugging() >= 1 then
+        local display = {}
+        local limit = 0
+        for co, thread in pairs(running) do
+          local this = display[thread.short_basename]
+          if not this then
+            this = {}
+            limit = limit + 1
+            if limit > 5 then
+              -- Only print stats if 5 or fewer scripts remaining
+              break
+            end
+          end
+          this[1] = (this[1] or 0) + 1
+          display[thread.short_basename] = this
+        end
+        for co, thread in pairs(waiting) do
+          local this = display[thread.short_basename]
+          if not this then
+            this = {}
+            limit = limit + 1
+            if limit > 5 then
+              -- Only print stats if 5 or fewer scripts remaining
+              break
+            end
+          end
+          this[2] = (this[2] or 0) + 1
+          display[thread.short_basename] = this
+        end
+        if limit <= 5 then
+          for name, stats in pairs(display) do
+            print_debug(1, "Script %s: %d threads running, %d threads waiting",
+              name, stats[1] or 0, stats[2] or 0)
+          end
+        end
       end
-    elseif progress "mayBePrinted" then
+    elseif total > 0 and progress "mayBePrinted" then
       if verbosity() > 1 or debugging() > 0 then
         progress("printStats", 1-(nr+nw)/total);
       else
@@ -1165,7 +1220,7 @@ do
     local t, path = cnse.fetchfile_absolute(cnse.scriptargsfile)
     assert(t == 'file', format("%s is not a file", path))
     print_debug(1, "Loading script-args from file `%s'", cnse.scriptargsfile);
-    args[#args+1] = assert(assert(open(path, 'r')):read "*a"):gsub("\n", ","):gsub(",*$", "");
+    args[#args+1] = assert(assert(open(path, 'r')):read "*a"):gsub(",*$", "");
   end
 
   if cnse.scriptargs then -- Load script arguments (--script-args)
@@ -1173,6 +1228,10 @@ do
     args[#args+1] = cnse.scriptargs;
   end
 
+  if cnse.script_timeout and cnse.script_timeout > 0 then
+    print_debug(1, "Set script-timeout as: %d seconds", cnse.script_timeout);
+  end
+
   args = concat(args, ",");
   if #args > 0 then
     print_debug(1, "Arguments parsed: %s", args);
@@ -1187,15 +1246,17 @@ do
     local parser = locale {
       V "space"^0 * V "table" * V "space"^0,
       table = Cf(Ct "" * P "{" * V "space"^0 * (V "fieldlst")^-1 * V "space"^0 * P "}", set);
-      fieldlst = V "field" * (V "space"^0 * P "," * V "space"^0 * V "field")^0;
+      hws = V "space" - P "\n", -- horizontal whitespace
+      fieldlst = V "field" * (V "hws"^0 * S "\n," * V "space"^0 * V "field")^0;
       field = V "kv" + V "av";
-      kv = Cg(V "string" * V "space"^0 * P "=" * V "space"^0 * V "value");
+      kv = Cg(V "string" * V "hws"^0 * P "=" * V "hws"^0 * V "value");
       av = Cg(V "value");
       value = V "table" + V "string";
       string = V "qstring" + V "uqstring";
       qstring = U.escaped_quote('"') + U.escaped_quote("'");
-      uqstring = V "space"^0 * C((P(1) - V "space"^0 * S ",{}=")^0) * V "space"^0; -- everything but ',{}=', do not capture final space
+      uqstring = V "hws"^0 * C((P(1) - V "hws"^0 * S "\n,{}=")^0) * V "hws"^0; -- everything but '\n,{}=', do not capture final space
     };
+    --U.debug(parser,function(...)return print_debug(1,...)end)
     parser = assert(P(parser));
     nmap.registry.args = parser:match("{"..args.."}");
     if not nmap.registry.args then
@@ -1302,6 +1363,10 @@ local function main (hosts, scantype)
     insert(runlevels[script.runlevel], script);
   end
 
+  if _R[PARALLELISM] > CONCURRENCY_LIMIT then
+    CONCURRENCY_LIMIT = _R[PARALLELISM];
+  end
+
   if scantype == NSE_PRE_SCAN then
     print_verbose(1, "Script Pre-scanning.");
   elseif scantype == NSE_SCAN then
