@@ -11,9 +11,12 @@
 
 local stdnse = require "stdnse"
 local string = require "string"
+local stringaux = require "stringaux"
 local math = require "math"
 local os = require "os"
 local table = require "table"
+local tableaux = require "tableaux"
+local rand = require "rand"
 _ENV = stdnse.module("tls", stdnse.seeall)
 
 local pack = string.pack
@@ -675,15 +678,8 @@ DEFAULT_CIPHERS = {
 }
 
 local function find_key(t, value)
-  local k, v
-
-  for k, v in pairs(t) do
-    if v == value then
-      return k
-    end
-  end
-
-  return nil
+  local found, v = tableaux.contains(t, value)
+  return v
 end
 
 -- Keep this local to enforce use of the cipher_info function
@@ -897,6 +893,7 @@ KEX_ALGORITHMS.RSA_EXPORT1024 = KEX_ALGORITHMS.RSA_EXPORT
 KEX_ALGORITHMS.DHE_RSA={
   pubkey="rsa",
   type = "dh",
+  pfs = true,
   server_key_exchange = function (blob, protocol)
     local pos
     local ret = {}
@@ -909,17 +906,20 @@ KEX_ALGORITHMS.DHE_RSA_EXPORT={
   export=true,
   pubkey="rsa",
   type = "dh",
+  pfs = true,
   server_key_exchange = KEX_ALGORITHMS.DHE_RSA.server_key_exchange
 }
 KEX_ALGORITHMS.DHE_DSS={
   pubkey="dsa",
   type = "dh",
+  pfs = true,
   server_key_exchange = KEX_ALGORITHMS.DHE_RSA.server_key_exchange
 }
 KEX_ALGORITHMS.DHE_DSS_EXPORT={
   export=true,
   pubkey="dsa",
   type = "dh",
+  pfs = true,
   server_key_exchange = KEX_ALGORITHMS.DHE_RSA.server_key_exchange
 }
 KEX_ALGORITHMS.DHE_DSS_EXPORT1024 = KEX_ALGORITHMS.DHE_DSS_EXPORT1024
@@ -942,6 +942,7 @@ KEX_ALGORITHMS.DH_RSA_EXPORT={
 KEX_ALGORITHMS.ECDHE_RSA={
   pubkey="rsa",
   type = "ec",
+  pfs = true,
   server_key_exchange = function (blob, protocol)
     local pos
     local ret = {}
@@ -953,6 +954,7 @@ KEX_ALGORITHMS.ECDHE_RSA={
 KEX_ALGORITHMS.ECDHE_ECDSA={
   pubkey="ec",
   type = "ec",
+  pfs = true,
   server_key_exchange = KEX_ALGORITHMS.ECDHE_RSA.server_key_exchange
 }
 KEX_ALGORITHMS.ECDH_ECDSA={
@@ -995,6 +997,7 @@ KEX_ALGORITHMS.RSA_PSK = {
 }
 KEX_ALGORITHMS.DHE_PSK = {
   type = "dh",
+  pfs = true,
   server_key_exchange = function (blob, protocol)
     local pos
     local ret = {}
@@ -1009,6 +1012,7 @@ KEX_ALGORITHMS.PSK_DHE = KEX_ALGORITHMS.DHE_PSK
 --rfc5489
 KEX_ALGORITHMS.ECDHE_PSK={
   type = "ec",
+  pfs = true,
   server_key_exchange = function (blob, protocol)
     local pos
     local ret = {}
@@ -1021,6 +1025,7 @@ KEX_ALGORITHMS.ECDHE_PSK={
 -- RFC 5054
 KEX_ALGORITHMS.SRP_SHA = {
   type = "srp",
+  pfs = true,
   server_key_exchange = function (blob, protocol)
     local pos
     local ret = {srp={}}
@@ -1033,11 +1038,13 @@ KEX_ALGORITHMS.SRP_SHA = {
 KEX_ALGORITHMS.SRP_SHA_DSS = {
   pubkey="dsa",
   type = "srp",
+  pfs = true,
   server_key_exchange = KEX_ALGORITHMS.SRP_SHA.server_key_exchange
 }
 KEX_ALGORITHMS.SRP_SHA_RSA = {
   pubkey="rsa",
   type = "srp",
+  pfs = true,
   server_key_exchange = KEX_ALGORITHMS.SRP_SHA.server_key_exchange
 }
 
@@ -1067,7 +1074,7 @@ function cipher_info (c)
   local info = cipher_info_cache[c]
   if info then return info end
   info = {}
-  local tokens = stdnse.strsplit("_", c)
+  local tokens = stringaux.strsplit("_", c)
   local i = 1
   if tokens[i] ~= "TLS" and tokens[i] ~= "SSL" then
     stdnse.debug2("cipher_info: Not a TLS ciphersuite: %s", c)
@@ -1205,7 +1212,13 @@ handshake_parse = {
         end
         local b = {certificates = {}}
         while j < cert_end do
-          local cert_len, cert
+          local cert_len = unpack(">I3", buffer, j)
+          if cert_len + 3 + j > cert_end then
+            stdnse.debug1("server_certificate parsing error!")
+            j = cert_end
+            break
+          end
+          local cert
           cert, j = unpack(">s3", buffer, j)
           -- parse these with sslcert.parse_ssl_certificate
           table.insert(b["certificates"], cert)
@@ -1448,7 +1461,7 @@ function client_hello(t)
   ))
 
   -- Set the random data.
-  table.insert(b, stdnse.generate_random_string(28))
+  table.insert(b, rand.random_string(28))
 
   -- Set the session ID.
   local sid = t["session_id"] or ""
@@ -1527,7 +1540,15 @@ function client_hello(t)
   table.insert(h, pack(">s3", b))
 
   -- Record layer version should be SSLv3 (lowest compatible record version)
-  return record_write("handshake", t.record_protocol or "SSLv3", table.concat(h))
+  -- But some implementations (OpenSSL) will not finish a handshake that could
+  -- be downgraded by a MITM to SSLv3. So we use TLSv1.0 unless the caller
+  -- explicitly tries to set SSLv3.0 somewhere (t.record_protocol or
+  -- t.protocol)
+  local record_proto = t.record_protocol
+  if not record_proto then
+    record_proto = (t.protocol == "SSLv3") and "SSLv3" or "TLSv1.0"
+  end
+  return record_write("handshake", record_proto, table.concat(h))
 end
 
 local function read_atleast(s, n)
